#!/usr/bin/env python3
#
#  __init__.py
"""
Supported sources for the requirements are implemented here.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Redistribution and use in source and binary forms, with or without modification,
#  are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright notice,
#        this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import importlib.util
import inspect
import mimetypes
import pathlib
from typing import Any, Callable, Dict, List, Tuple

# 3rd party
import sphinx.environment
from docutils.parsers.rst import directives
from domdf_python_tools.paths import PathPlus
from setuptools.config import read_configuration  # type: ignore

# this package
from sphinxcontrib.extras_require.flit_config import read_flit_config

__all__ = [
		"requirements_from_file",
		"requirements_from___pkginfo__",
		"requirements_from_setup_cfg",
		"requirements_from_flit",
		"flag",
		"sources",
		"Sources",
		]


class Sources(List[Tuple[str, Callable, Callable]]):
	"""
	Class to store functions that provide requirements sources.

	The syntax of each entry is:

	``(option_name, getter_function, validator_function)``

	* a string to use in the directive to specify the source to use,
	* the function that returns the list of additional requirements,
	* a function to validate the option value provided by the user.
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

	_args = ["package_root", "options", "env", "extra"]
	_directive_name = "extras_require"

	def register(
			self,
			option_name: str,
			validator: Callable = directives.unchanged,
			) -> Callable:
		"""
		Decorator to register a function.

		The function must have the following signature:

		.. code-block:: python

			def function(
				package_root: pathlib.Path,
				options: Dict,
				env: sphinx.environment.BuildEnvironment,
				extra: str,
				) -> List[str]: ...

		:param option_name: A string to use in the directive to specify the source to use.
		:param validator: A function to validate the option value provided by the user.

		:return: The registered function.

		:raises: :exc:`SyntaxError` if the decorated function does not take the correct arguments.
		"""

		def _decorator(function: Callable) -> Callable:
			signature = inspect.signature(function)

			if list(signature.parameters.keys()) != self._args:
				raise SyntaxError(
						"The decorated function must take only the following arguments: "
						"'package_root', 'options', 'env', and 'extra'"
						)

			self.append((option_name, function, validator))

			setattr(function, f"_{self._directive_name}_registered", True)

			return function

		return _decorator


#: Instance of :class:`~.Sources`.
sources = Sources()


def flag(argument: Any) -> bool:
	"""
	Check for a valid flag option (no argument) and return :py:obj:`True`.

	:raises: :exc:`ValueError` if an argument is given.
	"""

	if argument and argument.strip():
		raise ValueError(f'No argument is allowed; "{argument}" supplied')
	else:
		return True


@sources.register("file", directives.unchanged)
def requirements_from_file(
		package_root: pathlib.Path,
		options: Dict,
		env: sphinx.environment.BuildEnvironment,
		extra: str,
		) -> List[str]:
	"""
	Load requirements from the specified file.

	:param package_root: The path to the package root
	:param options:
	:param env:
	:param extra: The name of the "extra" that the requirements are for

	:return: List of requirements
	"""

	requirements_file = package_root / options["file"]

	if not requirements_file.is_file():
		raise FileNotFoundError(f"Cannot find requirements file '{requirements_file}'")

	mime_type = mimetypes.guess_type(str(requirements_file))[0]
	if not mime_type or not mime_type.startswith("text/"):
		raise ValueError(f"'{requirements_file}' is not a text file.")

	requirements = [x for x in requirements_file.read_text().split("\n") if x and not x.startswith("#")]

	return requirements


@sources.register("__pkginfo__", flag)
def requirements_from___pkginfo__(
		package_root: pathlib.Path,
		options: Dict,
		env: sphinx.environment.BuildEnvironment,
		extra: str,
		) -> List[str]:
	"""
	Load requirements from a ``__pkginfo__.py`` file in the root of the repository.

	:param package_root: The path to the package root
	:param options:
	:param env:
	:param extra: The name of the "extra" that the requirements are for

	:return: List of requirements
	"""

	__pkginfo___file = PathPlus(env.srcdir).parent / "__pkginfo__.py"

	if not __pkginfo___file.is_file():
		raise FileNotFoundError(f"Cannot find __pkginfo__.py in '{__pkginfo___file.parent}'")

	try:
		spec = importlib.util.spec_from_file_location("__pkginfo__", str(__pkginfo___file))

		if spec is not None:
			__pkginfo__ = importlib.util.module_from_spec(spec)

			if spec.loader:
				spec.loader.exec_module(__pkginfo__)  # type: ignore
				requirements = __pkginfo__.extras_require[extra]  # type: ignore
				return requirements
				# TODO: handle extra not found

	except ValueError:
		pass

	raise ImportError("Could not import __pkginfo__.py")


@sources.register("setup.cfg", flag)
def requirements_from_setup_cfg(
		package_root: pathlib.Path,
		options: Dict,
		env: sphinx.environment.BuildEnvironment,
		extra: str,
		) -> List[str]:
	"""
	Load requirements from a ``setup.cfg`` file in the root of the repository.

	:param package_root: The path to the package root
	:param options:
	:param env:
	:param extra: The name of the "extra" that the requirements are for

	:return: List of requirements
	"""

	setup_cfg_file = PathPlus(env.srcdir).parent / "setup.cfg"
	assert setup_cfg_file.is_file()

	setup_cfg = read_configuration(setup_cfg_file)

	if "options" in setup_cfg and "extras_require" in setup_cfg["options"]:
		if extra in setup_cfg["options"]["extras_require"]:
			return setup_cfg["options"]["extras_require"][extra]
		else:
			raise ValueError(f"'{extra}' not found in '[options.extras_require]'")
	else:
		raise ValueError("'options.extras_require' section not found in 'setup.cfg")


@sources.register("flit", flag)
def requirements_from_flit(
		package_root: pathlib.Path,
		options: Dict,
		env: sphinx.environment.BuildEnvironment,
		extra: str,
		) -> List[str]:
	"""
	Load requirements from the ``[tool.flit.metadata.requires-extra]`` section of
	a pyproject.toml file in the root of the repository.

	:param package_root: The path to the package root
	:param options:
	:param env:
	:param extra: The name of the "extra" that the requirements are for

	:return: List of requirements
	"""  # noqa D400

	pyproject_file = PathPlus(env.srcdir).parent / "pyproject.toml"

	if not pyproject_file.is_file():
		raise FileNotFoundError(f"Cannot find pyproject.toml in '{pyproject_file.parent}'")

	flit_extras = read_flit_config(pyproject_file).reqs_by_extra

	if extra in flit_extras:
		return flit_extras[extra]
	else:
		raise ValueError(f"'{extra}' not found in '[tool.flit.metadata.requires-extra]'")



#!/usr/bin/env python3
#
#  __init__.py
"""
A Sphinx directive to specify that a module has extra requirements, and show how to install them.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Redistribution and use in source and binary forms, with or without modification,
#  are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright notice,
#        this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from typing import Any, Dict

# 3rd party
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

# this package
from sphinxcontrib.extras_require.directive import ExtrasRequireDirective
from sphinxcontrib.extras_require.sources import sources

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "BSD"
__version__: str = "0.2.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = ["purge_extras_requires", "setup"]

# For type hinting install docutils-stubs


def purge_extras_requires(app: Sphinx, env: BuildEnvironment, docname: str) -> None:
	"""
	Remove all redundant :class:`extras_require.directive.ExtrasRequireDirective` nodes.

	:param app:
	:param env:
	:param docname: The name of the document to remove nodes for.
	"""

	if not hasattr(env, "all_extras_requires"):
		return

	env.all_extras_requires = [  # type: ignore
		todo for todo in env.all_extras_requires  # type: ignore
		if todo["docname"] != docname
		]


def setup(app: Sphinx) -> Dict[str, Any]:
	"""
	Setup Sphinx Extension.

	:param app:

	:return:
	"""

	app.setup_extension("sphinx-prompt")

	# Location of package source directory relative to documentation source directory
	app.add_config_value("package_root", None, "html", [str])

	app.add_directive("extras-require", ExtrasRequireDirective)
	app.connect("env-purge-doc", purge_extras_requires)

	return {
			"version": __version__,
			"parallel_read_safe": True,
			"parallel_write_safe": True,
			}



# From https://github.com/takluyver/flit/blob/master/flit_core/flit_core/config.py
#
# Copyright (c) 2015, Thomas Kluyver and contributors
# All rights reserved.
#
# BSD 3-clause license:
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# stdlib
import difflib
import logging

# 3rd party
import pytoml as toml  # type: ignore

__all__ = ["ConfigError", "read_flit_config", "prep_toml_config", "LoadedConfig"]

log = logging.getLogger(__name__)


class ConfigError(ValueError):
	pass


metadata_list_fields = {"classifiers", "requires", "dev-requires"}

metadata_allowed_fields = {
		"module",
		"author",
		"author-email",
		"maintainer",
		"maintainer-email",
		"home-page",
		"license",
		"keywords",
		"requires-python",
		"dist-name",
		"description-file",
		"requires-extra",
		} | metadata_list_fields

metadata_required_fields = {
		"module",
		"author",
		}


def read_flit_config(path):
	"""Read and check the `pyproject.toml` file with data about the package.
	"""
	with path.open('r', encoding="utf-8") as f:
		d = toml.load(f)
	return prep_toml_config(d, path)


def prep_toml_config(d, path):
	"""Validate config loaded from pyproject.toml and prepare common metadata

	Returns a LoadedConfig object.
	"""
	if ("tool" not in d) or ("flit" not in d["tool"]) or (not isinstance(d["tool"]["flit"], dict)):
		raise ConfigError("TOML file missing [tool.flit] table.")

	d = d["tool"]["flit"]
	unknown_sections_ = set(d) - {"metadata", "scripts", "entrypoints", "sdist"}
	unknown_sections = [s for s in unknown_sections_ if not s.lower().startswith("x-")]
	if unknown_sections:
		raise ConfigError("Unknown sections: " + ", ".join(unknown_sections))

	if "metadata" not in d:
		raise ConfigError("[tool.flit.metadata] section is required")

	loaded_cfg = _prep_metadata(d["metadata"], path)

	return loaded_cfg


class LoadedConfig:

	def __init__(self):
		self.module = None
		self.metadata = {}
		self.reqs_by_extra = {}
		self.entrypoints = {}
		self.referenced_files = []
		self.sdist_include_patterns = []
		self.sdist_exclude_patterns = []


readme_ext_to_content_type = {
		".rst": "text/x-rst",
		".md": "text/markdown",
		".txt": "text/plain",
		}


def _prep_metadata(md_sect, path):
	"""Process & verify the metadata from a config file

	- Pull out the module name we're packaging.
	- Read description-file and check that it's valid rst
	- Convert dashes in key names to underscores
	  (e.g. home-page in config -> home_page in metadata)
	"""
	if not set(md_sect).issuperset(metadata_required_fields):
		missing = metadata_required_fields - set(md_sect)
		raise ConfigError("Required fields missing: " + '\n'.join(missing))

	res = LoadedConfig()

	res.module = md_sect.get("module")
	if not str.isidentifier(res.module):
		raise ConfigError(f"Module name {res.module!r} is not a valid identifier")

	md_dict = res.metadata

	for key, value in md_sect.items():
		if key in ["description-file", "module"]:
			continue
		if key not in metadata_allowed_fields:
			closest = difflib.get_close_matches(key, metadata_allowed_fields, n=1, cutoff=0.7)
			msg = f"Unrecognised metadata key: {key!r}"
			if closest:
				msg += f" (did you mean {closest[0]!r}?)"
			raise ConfigError(msg)

		k2 = key.replace('-', '_')
		md_dict[k2] = value
		if key == "requires-extra":
			if not isinstance(value, dict):  # pragma: no cover
				raise ConfigError(f"Expected a dict for requires-extra field, found {value!r}")
			if not all(isinstance(e, list) for e in value.values()):
				raise ConfigError("Expected a dict of lists for requires-extra field")
			for e, reqs in value.items():
				if not all(isinstance(a, str) for a in reqs):
					raise ConfigError(f"Expected a string list for requires-extra. (extra {e})")
		else:
			if not isinstance(value, str):
				raise ConfigError(f"Expected a string for {key} field, found {value!r}")

	# Move dev-requires into requires-extra
	reqs_noextra = md_dict.pop("requires_dist", [])
	res.reqs_by_extra = md_dict.pop("requires_extra", {})
	dev_requires = md_dict.pop("dev_requires", None)
	if dev_requires is not None:
		if "dev" in res.reqs_by_extra:
			raise ConfigError("dev-requires occurs together with its replacement requires-extra.dev.")
		else:
			log.warning('"dev-requires = ..." is obsolete. Use "requires-extra = {"dev" = ...}" instead.')
			res.reqs_by_extra["dev"] = dev_requires

	# Add requires-extra requirements into requires_dist
	md_dict["requires_dist"] = reqs_noextra + list(_expand_requires_extra(res.reqs_by_extra))

	md_dict["provides_extra"] = sorted(res.reqs_by_extra.keys())

	# For internal use, record the main requirements as a '.none' extra.
	res.reqs_by_extra[".none"] = reqs_noextra

	return res


def _expand_requires_extra(re):
	for extra, reqs in sorted(re.items()):
		for req in reqs:
			if ';' in req:
				name, envmark = req.split(';', 1)
				yield f'{name} ; extra == "{extra}" and ({envmark})'
			else:
				yield f'{req} ; extra == "{extra}"'



#!/usr/bin/env python3
#
#  __init__.py
"""
The "extras_require" directive.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Redistribution and use in source and binary forms, with or without modification,
#  are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright notice,
#        this list of conditions and the following disclaimer in the documentation
#        and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
#  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import textwrap
import warnings
from typing import Any, Dict, Iterable, List, Union

# 3rd party
from docutils import nodes
from docutils.statemachine import ViewList
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.stringlist import StringList
from packaging.requirements import InvalidRequirement, Requirement
from sphinx.util.docutils import SphinxDirective

# this package
from sphinxcontrib.extras_require.sources import sources

__all__ = ["ExtrasRequireDirective", "validate_requirements", "make_node_content", "get_requirements"]


class ExtrasRequireDirective(SphinxDirective):
	"""
	Directive to show a notice to users that a module, class or function has additional requirements.
	"""

	has_content: bool = True
	required_arguments: int = 1
	option_spec = {source[0]: source[2] for source in sources}  # type: ignore
	option_spec["scope"] = str

	def run(self) -> List[nodes.Node]:
		"""
		Create the extras_require node.

		:return:
		"""

		extra: str = self.arguments[0]

		targetid = f'extras_require-{self.env.new_serialno("extras_require"):d}'
		targetnode = nodes.target('', '', ids=[targetid])

		valid_requirements = get_requirements(
				env=self.env,
				extra=extra,
				options=self.options,
				content=self.content,
				)

		if not valid_requirements:
			warnings.warn("No requirements specified! No notice will be shown in the documentation.")
			return [targetnode]

		scope = self.options.get("scope", "module")

		content = make_node_content(valid_requirements, self.env.config.project, extra, scope=scope)
		view = ViewList(content.split("\n"))

		extras_require_node = nodes.attention(rawsource=content)
		self.state.nested_parse(view, self.content_offset, extras_require_node)  # type: ignore

		if not hasattr(self.env, "all_extras_requires"):
			self.env.all_extras_requires = []  # type: ignore

		self.env.all_extras_requires.append({  # type: ignore
			"docname": self.env.docname,
			"lineno": self.lineno,
			"extras_require": extras_require_node.deepcopy(),
			"target": targetnode,
			})

		return [targetnode, extras_require_node]


def validate_requirements(requirements_list: List[str]) -> List[str]:
	"""
	Validate a list of :pep:`508` requirements and format them consistently.

	:param requirements_list: List of :pep:`508` requirements.

	:return: List of :pep:`508` requirements with consistent formatting.
	"""

	valid_requirements = []

	for req in requirements_list:
		if req:
			try:
				valid_requirements.append(Requirement(req))
			except InvalidRequirement as e:
				raise ValueError(f"Invalid requirement '{req}': {str(e)}") from None

	valid_requirements.sort(key=lambda r: r.name)

	return [str(x) for x in valid_requirements]


def make_node_content(
		requirements: List[str],
		package_name: str,
		extra: str,
		scope: str = "module",
		) -> str:
	"""
	Create the content of an extras_require node.

	:param requirements: List of additional :pep:`508` requirements.
	:param package_name: The name of the module/package on PyPI.
	:type package_name: str
	:param extra: The name of the "extra".
	:type extra: str
	:param scope: The scope of the additional requirements, e.g. ``"module"``, ``"package"``.
	:type scope: str

	:return: The content of an extras_require node.
	:rtype:
	"""

	if len(requirements) > 1:
		plural = 's'
	else:
		plural = ''

	content = StringList(convert_indents=True)
	content.indent_type = " " * 4
	content.append(f"This {scope} has the following additional requirement{plural}:")
	content.blankline(ensure_single=True)

	with content.with_indent_size(content.indent_size + 1):
		content.append(".. code-block:: text")
		content.blankline(ensure_single=True)

		with content.with_indent_size(content.indent_size + 1):
			content.extend(requirements)

	content.blankline(ensure_single=True)

	content.append("These can be installed as follows:")
	content.blankline(ensure_single=True)

	with content.with_indent_size(content.indent_size + 1):
		content.append(".. prompt:: bash")
		content.blankline(ensure_single=True)

		with content.with_indent_size(content.indent_size + 1):
			content.append(f"python -m pip install {package_name}[{extra}]")

	content.blankline(ensure_single=True)
	content.blankline()

	return str(content)


def get_requirements(env, extra: str, options: Dict[str, Any], content: Union[Iterable, ViewList]) -> List[str]:
	"""
	Get the requirements for the extras_require node.

	:param env:
	:type env:
	:param extra:
	:type extra: str
	:param options:
	:param content:

	:return:
	"""

	n_sources = 0
	if list(content):
		n_sources += 1
	for source in sources:
		if (source[0] in options) and options[source[0]]:
			n_sources += 1

	if n_sources > 1:
		raise ValueError("Please specify only one source for the extra requirements")
	elif n_sources == 0:
		raise ValueError(f"Please specify a source for the extra requirements {extra}")

	src_dir = PathPlus(env.srcdir)
	package_root = src_dir.parent / env.config.package_root

	requirements: List[str]

	for option_name, getter_function, validator_function in sources:

		if option_name in options:
			requirements = getter_function(package_root, options, env, extra)
			break
	else:
		requirements = list(content)

	valid_requirements = validate_requirements(requirements)

	return valid_requirements
